import os
import shutil
from typing import Iterator

import fire
from colorama import Fore, deinit, init
from psycopg2 import ProgrammingError, connect, sql
from psycopg2.extras import RealDictCursor
from requests import HTTPError

from fhir_cli import (
    DBT_META_TABLE,
    DBT_SCHEMA,
    FHIR_COLUMN_NAME,
    FHIR_DBT_SCHEMA,
    PACKAGE_PATH,
    POSTGRES_DB,
    POSTGRES_HOST,
    POSTGRES_PASSWORD,
    POSTGRES_PORT,
    POSTGRES_USER,
    log,
)
from fhir_cli.admin import Admin
from fhir_cli.dbt import Dbt
from fhir_cli.fhir_resource import FhirResource, FhirValidationError
from utils.compact import dict_compact
from utils.dotty import dotty_dict
from utils.number_print import number_print

CURSOR_ITERSIZE = 20


def get_fhir_resources_from_model(
    cursor, model: str, offset: int = 0, limit: int = 100
) -> Iterator[dict]:
    """get_fhir_resources_from_model looks for a fhir model file and retrieves the Fhir resources

    This function uses a named cursor fetching at most CURSOR_ITERSIZE rows
    at each network roundtrip during iteration on the cursor
    https://www.psycopg.org/docs/cursor.html#cursor.itersize

    Args:
        cursor: a database connection cursor
        model (str): a Fhir model name
        offset (:obj:`int`, optional): an offset for the executed query. Defaults to 0.
        limit (:obj:`int`, optional): a limit for the executed query. Defaults to 100.
    """
    select_fhir_stmt = sql.SQL(
        "SELECT {fhir_column_name} FROM {fhir_model} LIMIT %s OFFSET %s"
    ).format(
        fhir_column_name=sql.Identifier(FHIR_COLUMN_NAME),
        fhir_model=sql.Identifier(model),
    )
    cursor.execute(select_fhir_stmt, (limit, offset))
    for row in cursor:
        yield dotty_dict(dict_compact(row[FHIR_COLUMN_NAME]))


class Cli:
    """a cli to manage your DbtOnFhir project"""

    def __init__(self):
        self.dbt = Dbt()
        self.admin = Admin()

    @staticmethod
    def init(omop=False):
        """Generate a new project

        Args:
            omop (:obj:`bool`, optional): include OMOP DBT models
        """

        shutil.copytree(
            os.path.join(PACKAGE_PATH, "starter"),
            ".",
            ignore=shutil.ignore_patterns("*omop") if not omop else None,
            dirs_exist_ok=True,
        )

    @staticmethod
    def add_omop_models():
        """Add OMOP DBT models"""

        shutil.copytree(os.path.join(PACKAGE_PATH, "starter", "schemas/omop"), "./schemas/omop")

    @staticmethod
    def link(database: str, schema: str, include: tuple[str] = None, exclude: tuple[str] = None):
        """Create a foreign schema from a schema of the source database

        This command uses a foreign data wrapper to access data
        stored in an external server. The module opens a connection and extracts
        a database schema. Subsequently a user can import the data to the local
        database by creating a table or a materialized view selecting from a
        foreign table of a given schema.

        Args:
            database (str): the source database
            schema (str): the source schema
            include (:obj:`tuple` of :obj:`str`, optional): comma separated table names to import
            only foreign tables matching one of the given table names. Other tables existing in
            the foreign schema will be ignored.
            exclude (:obj:`tuple` of :obj:`str`, optional): comma separated table names to exclude
            specified  foreign tables from the import. All tables existing in the foreign
            schema will be imported except the ones listed here.
        """

        if include and exclude:
            log.error("include and exclude options are mutually exclusive")
            return

        conn = connect(
            host=POSTGRES_HOST,
            dbname=POSTGRES_DB,
            port=POSTGRES_PORT,
            user=POSTGRES_USER,
            password=POSTGRES_PASSWORD,
        )

        try:
            with conn:
                with conn.cursor() as curs:
                    db_type_stmt = sql.SQL("SELECT db_type FROM {dbt_meta_table}").format(
                        dbt_meta_table=sql.Identifier(DBT_SCHEMA, DBT_META_TABLE)
                    )
                    curs.execute(db_type_stmt)
                    db_type = curs.fetchone()
        except Exception as e:
            conn.close()
            log.error(e)
            return

        schema_stmt = sql.SQL("CREATE SCHEMA IF NOT EXISTS {target_schema}").format(
            target_schema=sql.Identifier(f"{database}_{schema}")
        )

        constraints = sql.SQL("")
        if include:
            included_tables = list(map(lambda table: sql.Identifier(table), include))
            constraints = sql.SQL("LIMIT TO ({included_tables})").format(
                included_tables=sql.SQL(", ").join(included_tables)
            )
        elif exclude:
            excluded_tables = list(map(lambda table: sql.Identifier(table), exclude))
            constraints = sql.SQL("EXCEPT ({excluded_tables})").format(
                excluded_tables=sql.SQL(", ").join(excluded_tables)
            )

        # increase the number of rows prefetched from the oracle db to improve perf.
        # see https://github.com/laurenz/oracle_fdw#foreign-table-options
        options = sql.SQL("OPTIONS (prefetch '10240')") if db_type == "oracle" else sql.SQL("")

        import_stmt = sql.SQL(
            "IMPORT FOREIGN SCHEMA {source_schema} "
            "{constraints} "
            "FROM SERVER {server} INTO {target_schema} "
            "{options}"
        ).format(
            source_schema=sql.Identifier(schema),
            target_schema=sql.Identifier(f"{database}_{schema}"),
            constraints=constraints,
            server=sql.Identifier(f"source_server_{database}"),
            options=options,
        )

        try:
            with conn:
                with conn.cursor() as curs:
                    curs.execute(schema_stmt)
                    curs.execute(import_stmt)
        except Exception as e:
            log.error(e)
        finally:
            conn.close()

    @staticmethod
    def validate(model: str, offset: int = 0, limit: int = 100):
        """Extract a fhir model row and validates the Fhir resource
        against a Fhir server

        Args:
            model (str): should be a valid DBT Fhir model name such as `observation_heartrate`
            offset (:obj:`int`, optional): set an offset to the query. Defaults to 0.
            limit (:obj:`int`, optional): a limit for the executed query. Defaults to 100.
        """
        conn = connect(
            host=POSTGRES_HOST,
            dbname=POSTGRES_DB,
            port=POSTGRES_PORT,
            user=POSTGRES_USER,
            password=POSTGRES_PASSWORD,
            cursor_factory=RealDictCursor,
            options=f"-c search_path={FHIR_DBT_SCHEMA}",
        )
        cursor = conn.cursor(name="curname")
        try:
            # Function in order to filter ANSI escape sequences on Windows
            # see https://pypi.org/project/colorama/
            init(autoreset=True)
            with conn:
                with cursor:
                    cursor.itersize = CURSOR_ITERSIZE
                    for resource in get_fhir_resources_from_model(cursor, model, offset, limit):
                        fhir = FhirResource(resource)
                        number_print(repr(fhir))
                        try:
                            fhir.validate()
                            # The message will be printed in GREEN, hence the weird sequence
                            # at the beginning and at the end
                            log.info(f"{Fore.GREEN}✅ FHIR resource is valid")
                        except FhirValidationError as e:
                            # Same as above
                            log.error(f"{Fore.RED}❌ FHIR resource is not valid\n{e}")
                        input("Press [Return] for the next result or [Ctrl+c] to quit")
        except (ProgrammingError, HTTPError) as e:
            log.error(e)
        except KeyboardInterrupt:
            pass
        finally:
            conn.close()
            # Stop using Colorama
            # see https://pypi.org/project/colorama/
            deinit()


def run():
    cli = Cli()
    fire.Fire(cli)


if __name__ == "__main__":
    run()



from dotty_dict import dotty


def dotty_list(lst: list) -> list:
    new_list = []
    for item in lst:
        if isinstance(item, dict):
            new_list.append(dotty_dict(item))
        elif isinstance(item, list):
            new_list.append(dotty_list(item))
        else:
            new_list.append(item)
    return new_list


def dotty_dict(obj: dict) -> dict:
    new_obj = {}
    for key, value in obj.items():
        data = dotty()
        if isinstance(value, dict):
            data[key] = dotty_dict(value)
        elif isinstance(value, list):
            data[key] = dotty_list(value)
        else:
            data[key] = value
        new_obj = new_obj | data.to_dict()
    return new_obj



def list_compact(array: list) -> list:
    for index, item in enumerate(array):
        if isinstance(item, dict):
            array[index] = dict_compact(item)
            if not len(array[index]):
                array[index] = None
        elif isinstance(item, list):
            array[index] = list_compact(item)
            if not len(array[index]):
                array[index] = None
    return [item for item in array if item is not None]


def dict_compact(obj: dict) -> dict:
    for key, value in obj.items():
        if isinstance(value, dict):
            obj[key] = dict_compact(value)
            if not len(obj[key]):
                obj[key] = None
        elif isinstance(value, list):
            obj[key] = list_compact(value)
            if not len(obj[key]):
                obj[key] = None
    return {key: value for key, value in obj.items() if value is not None}



import logging
import os

from dotenv import load_dotenv
from jinja2 import Environment, PackageLoader, select_autoescape

PACKAGE_PATH = os.path.dirname(__file__)

load_dotenv(override=True)

logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.INFO)
log = logging.getLogger(__name__)

JINJA_ENV = Environment(loader=PackageLoader("fhir_cli"), autoescape=select_autoescape())

# FHIR

FHIR_API_URL = os.environ.get("FHIR_API_URL", "http://localhost:8080/fhir")
FHIR_API_USER = os.environ.get("FHIR_API_USER")
FHIR_API_PASSWORD = os.environ.get("FHIR_API_PASSWORD")
FHIR_COLUMN_NAME = "fhir"

# OMOP

OMOP_TEMPLATES = (
    "01-OMOPCDM_postgresql_5.4_ddl.sql.j2",
    "02-OMOPCDM_postgresql_5.4_primary_keys.sql.j2",
    "03-vocabulary.sql.j2",
    "04-OMOPCDM_postgresql_5.4_constraints.sql.j2",
    "05-OMOPCDM_postgresql_5.4_indices.sql.j2",
)

# DBT

DBT_SCHEMA = os.environ.get("DBT_SCHEMA", "dbt")
FHIR_DBT_SCHEMA = f"{DBT_SCHEMA}_fhir"
OMOP_DBT_SCHEMA = f"{DBT_SCHEMA}_omop"
DBT_MODELS_DIR = "models"
DBT_META_TABLE = "_meta"
INIT_DB_TEMPLATE = "init_db.sql.j2"
MAPPING_DIR_NAME = os.environ.get("MAPPING_DIR_NAME", "schemas")

# Postgres

POSTGRES_HOST = os.environ.get("POSTGRES_HOST", "localhost")
POSTGRES_PORT = int(os.environ.get("POSTGRES_PORT", 5432))
POSTGRES_DB = os.environ.get("POSTGRES_DB", "postgres")
POSTGRES_USER = os.environ.get("POSTGRES_USER", "postgres")
POSTGRES_PASSWORD = os.environ.get("POSTGRES_PASSWORD", "secret")
POSTGRES_SERVER_NAME = os.environ.get("POSTGRES_SERVER_NAME", "postgres")

# Functions

FUNCTIONS_DIR_NAME = "functions"

# Connect

CONNECT_URL = os.environ.get("CONNECT_URL", "http://localhost:8083")
POSTGRES_SOURCE_CONNECTOR_TEMPLATE = "postgres_source_connector.json.j2"



import getpass
import glob
import re
from pathlib import Path
from typing import Iterator
from uuid import NAMESPACE_URL, uuid5

import requests
from psycopg2 import connect, sql

from fhir_cli import (
    CONNECT_URL,
    DBT_META_TABLE,
    DBT_SCHEMA,
    FHIR_DBT_SCHEMA,
    FUNCTIONS_DIR_NAME,
    INIT_DB_TEMPLATE,
    JINJA_ENV,
    OMOP_DBT_SCHEMA,
    OMOP_TEMPLATES,
    POSTGRES_DB,
    POSTGRES_HOST,
    POSTGRES_PASSWORD,
    POSTGRES_PORT,
    POSTGRES_SERVER_NAME,
    POSTGRES_SOURCE_CONNECTOR_TEMPLATE,
    POSTGRES_USER,
    log,
)


class InvalidJDBCUrlError(ValueError):
    pass


def get_user_defined_functions() -> Iterator[str]:
    """Yield SQL functions from the functions folder

    Returns:
        Iterator[str]: the SQL functions
    Raises:
        FileNotFoundError: if the function folder does not exist on the current path
    """
    if not Path(FUNCTIONS_DIR_NAME).is_dir():
        raise FileNotFoundError(f"{FUNCTIONS_DIR_NAME} folder not found")
    for file_path in glob.iglob(f"{FUNCTIONS_DIR_NAME}/**/*.sql", recursive=True):
        with open(file_path, "r") as f:
            yield f.read()


def db_type_from(url: str) -> str:
    """Extract the database type from the provided JDBC connection url

    Returns:
        str: the database type which can be either oracle or postgres
    Raises:
        InvalidJDBCUrlError: if the JDBC url is invalid
    """
    if (m := re.match(r"^jdbc:(postgres|oracle)://.+:\d+/\w+$", url)) is not None:
        return m.group(1)
    raise InvalidJDBCUrlError(f"JDBC url: {url} is invalid")


class Admin:
    """The admin command is used by an administrator to initialize a new project"""

    @staticmethod
    def createdb(database: str):
        """Create a new db for the project

        Args:
            database (str): a name for the new database
        """

        conn = connect(
            host=POSTGRES_HOST,
            dbname=POSTGRES_DB,
            port=POSTGRES_PORT,
            user=POSTGRES_USER,
            password=POSTGRES_PASSWORD,
        )
        conn.autocommit = True

        stmt = sql.SQL("CREATE DATABASE {database}").format(database=sql.Identifier(database))

        try:
            with conn.cursor() as curs:
                curs.execute(stmt)
        except Exception as e:
            log.error(e)
        finally:
            conn.close()

    @staticmethod
    def initdb(database: str, url: str, user: str = None, password: str = None, omop=False):
        """Initialize the new db for the project

        The omop vocabulary tables are populated during this process. It may take some time.

        Args:
            database (str): the database to initialize
            url (str): the source database JDBC url (formatted \
as jdbc:database_type//host:port/dbname) used to generate the project id
            user (:obj:`str`, optional): a username of choice.
            password (:obj:`str`, optional): a password of choice. If not specified, \
the command will prompt for a password
            omop (:obj:`bool`, optional): If specified, The omop vocabulary tables will be \
            populated during this process. It may take some time.
        """

        try:
            db_type = db_type_from(url)
        except InvalidJDBCUrlError as e:
            log.error(e)
            return

        while not password:
            password = getpass.getpass()

        conn = connect(
            host=POSTGRES_HOST,
            dbname=database,
            port=POSTGRES_PORT,
            user=POSTGRES_USER,
            password=POSTGRES_PASSWORD,
        )

        user_stmt = sql.SQL("CREATE USER {role} WITH PASSWORD %s").format(role=sql.Identifier(user))

        project_id = uuid5(NAMESPACE_URL, url)
        initdb_stmt = JINJA_ENV.get_template(INIT_DB_TEMPLATE).render(
            database=database,
            db_type=db_type,
            dbt_schema=DBT_SCHEMA,
            dbt_meta_table=DBT_META_TABLE,
            project_id=project_id,
            role=user,
            project_url=url,
        )

        try:
            with conn:
                with conn.cursor() as curs:
                    log.info("Creating a new user...")
                    curs.execute(user_stmt, (password,))
                    log.info("Initializing the project database...")
                    curs.execute(initdb_stmt)
                    if omop:
                        log.info("Loading OMOP vocabulary tables...")
                        for template in OMOP_TEMPLATES:
                            omop_stmt = JINJA_ENV.get_template(template).render(
                                schema=OMOP_DBT_SCHEMA
                            )
                            curs.execute(omop_stmt)
                    set_search_path_stmt = sql.SQL("SET search_path TO {dbt_schema}").format(
                        dbt_schema=sql.Identifier(DBT_SCHEMA)
                    )
                    curs.execute(set_search_path_stmt)
                    log.info("Loading user defined functions...")
                    for func_definition in get_user_defined_functions():
                        curs.execute(func_definition)
        except Exception as e:
            log.error(e)
        finally:
            conn.close()

    @staticmethod
    def add_omop_vocab(database: str):
        """Add omop vocabulary tables

        Args:
            database (str): the target database
        """
        conn = connect(
            host=POSTGRES_HOST,
            dbname=database,
            port=POSTGRES_PORT,
            user=POSTGRES_USER,
            password=POSTGRES_PASSWORD,
        )

        try:
            with conn:
                with conn.cursor() as curs:
                    log.info("Loading OMOP vocabulary tables...")
                    for template in OMOP_TEMPLATES:
                        omop_stmt = JINJA_ENV.get_template(template).render(schema=OMOP_DBT_SCHEMA)
                        curs.execute(omop_stmt)
        except Exception as e:
            log.error(e)
        finally:
            conn.close()

    @staticmethod
    def connect(database: str):
        """Add a Kafka Connect connector

        Args:
            database (str): the database to connect
        """
        source_connector = JINJA_ENV.get_template(POSTGRES_SOURCE_CONNECTOR_TEMPLATE).render(
            project_db=database,
            postgres_server_name=POSTGRES_SERVER_NAME,
            postgres_port=POSTGRES_PORT,
            postgres_user=POSTGRES_USER,
            postgres_password=POSTGRES_PASSWORD,
            fhir_schema=FHIR_DBT_SCHEMA,
        )
        r = requests.post(
            f"{CONNECT_URL}/connectors/",
            data=source_connector,
            headers={"Content-Type": "application/json"},
        )
        r.raise_for_status()

    @staticmethod
    def dblink(
        target: str, source: str, type: str, host: str, port: int, user: str, password: str = None
    ):
        """Create a foreign server linking the source database to the project database

        This command uses a foreign data wrapper to create a connection to another database
        thanks to the provided credentials.

        Args:
            target (str): the target database
            source (str): the source database
            type (str): the source database type (can be either `oracle`, `postgres` or `tds`)
            host (str): the source server host
            port (int): the source server port
            user (str): a source database user
            password (str): the user password. If not specified, the command will prompt for a
            password. If the password contains a special character, make sure you escape it properly
            `--password '"double#escape"'
        """

        while not password:
            password = getpass.getpass()

        conn = connect(
            host=POSTGRES_HOST,
            dbname=target,
            port=POSTGRES_PORT,
            user=POSTGRES_USER,
            password=POSTGRES_PASSWORD,
        )

        stmt = JINJA_ENV.get_template(f"{type}_fdw.sql.j2").render(
            source_db=source,
            source_host=host,
            source_port=port,
            source_user=user,
            source_password=password,
        )

        try:
            with conn:
                with conn.cursor() as curs:
                    curs.execute(stmt)
        except Exception as e:
            log.error(e)
        finally:
            conn.close()



import json

import requests

from fhir_cli import FHIR_API_PASSWORD, FHIR_API_URL, FHIR_API_USER
from utils.prettify_json import prettify_json


class FhirValidationError(ValueError):
    pass


class FhirResource:
    """FhirResource represents a fhir resource

    Args:
        resource (dict): dict containing the attributes of the resource
    """

    def __init__(self, resource: dict):
        self.resource = resource

    def __repr__(self):
        raw_json = json.dumps(self.resource, indent=4)
        return prettify_json(raw_json)

    def validate(self):
        """The validate method validates a Fhir resource against a Fhir server

        Raises:
            FhirValidationError: if the Fhir resource is invalid
            HttpError: if the request to the Fhir server returns an error
        """
        serialized_data = json.dumps(self.resource, indent=4)
        profile = None
        try:
            profile = self.resource["meta"]["profile"][0]
        except KeyError:
            pass
        try:
            resource_type = self.resource["resourceType"]
        except KeyError:
            raise FhirValidationError("resource has no resource type")
        r = requests.post(
            f"{FHIR_API_URL}/{resource_type}/$validate{f'?profile={profile}' if profile else ''}",
            data=serialized_data,
            headers={"Content-Type": "application/json+fhir"},
            auth=(FHIR_API_USER, FHIR_API_PASSWORD)
            if FHIR_API_USER and FHIR_API_PASSWORD
            else None,
        )
        if r.status_code == 404:
            raise FhirValidationError(f"{resource_type} is an unknown resource type")
        elif r.status_code != 400 and r.status_code != 412:
            r.raise_for_status()
        data = r.json()
        issues = (
            [issue for issue in data["issue"] if issue["severity"] == "error"]
            if "issue" in data
            else []
        )
        if len(issues):
            errors = ""
            for issue in issues:
                if issue.get("location"):
                    errors += (
                        f"* {issue['location'][0]}: {issue['diagnostics']}"
                        f" at {issue['location'][1]}\n"
                        if len(issue["location"]) == 2
                        else "\n"
                    )
                else:
                    errors += f"* {issue['diagnostics']}\n"
            raise FhirValidationError(f"the FHIR server returned a list of issues\n{errors}")



from pygments import highlight
from pygments.formatters.terminal256 import Terminal256Formatter
from pygments.lexers.data import JsonLexer


def prettify_json(raw_json: str):
    return highlight(
        raw_json,
        lexer=JsonLexer(),
        formatter=Terminal256Formatter(style="rrt"),
    )



import glob
import os
import re
import shutil
import subprocess  # nosec

from fhir_cli import DBT_MODELS_DIR, MAPPING_DIR_NAME, log


class UnknownDbtReferenceError(ValueError):
    pass


class EmptyDbtReferenceWarning(UserWarning):
    pass


def resolve_references(line: str) -> str:
    """resolve_references takes a line of a sql statement and looks for dbt references

    It looks for a comment such as `-- dbt-ref(reference_1, reference_2)` and replaces
    on the same line `reference_1` and `reference_2` by the jinja dbt references
    `{{ref('reference_1')}}` and `{{ref('reference_2')}}`

    Args:
        line (str): a line in the processed sql statement

    Returns:
        str: a new line with the references resolved or the input `line` if no reference
        have been found

    Raises:
        UnknownDbtReferenceError: if a reference is specified but not found in the line
        EmptyDbtReferenceWarning: if dbt-ref doesn't specify at least one reference
    """
    dbt_ref_pattern = r"(.*)(\s*--[ \t]*dbt-ref\((.*)\).*)"
    result = re.search(dbt_ref_pattern, line)
    if not result:
        return line
    stmt = result.group(1)
    references = [
        reference.strip() for reference in result.group(3).split(",") if reference.strip()
    ]
    if not references:
        raise EmptyDbtReferenceWarning("no reference has been specified")
    for reference in references:
        stmt, match = re.subn(rf"(\s+|^){reference}\b", f"\\1{{{{ref('{reference}')}}}}", stmt)
        if not match:
            raise UnknownDbtReferenceError(f"{reference} is unknown")
    return stmt


class Dbt:
    """The dbt command manages your DBT project"""

    @staticmethod
    def transpile():
        """Create the DBT models folder including all files in
        the schemas folder. It then compiles the sql files into DBT models by resolving
        DBT references and adding model configurations if needed.
        """
        if os.path.exists(DBT_MODELS_DIR):
            shutil.rmtree(DBT_MODELS_DIR)
        shutil.copytree(MAPPING_DIR_NAME, DBT_MODELS_DIR)

        for file_path in glob.iglob(f"{MAPPING_DIR_NAME}/**/*.sql", recursive=True):
            with open(file_path, "r") as f:
                with open(
                    f"{DBT_MODELS_DIR}{file_path.removeprefix(MAPPING_DIR_NAME)}", "w"
                ) as model_file:
                    output = []
                    for index, line in enumerate(f):
                        try:
                            new_line = resolve_references(line)
                        except UnknownDbtReferenceError as e:
                            log.error(f"model {file_path} line {index + 1}:\n{line}\nError: {e}")
                            shutil.rmtree(DBT_MODELS_DIR)
                            return
                        except EmptyDbtReferenceWarning as e:
                            log.warning(
                                f"model {file_path} line {index + 1}:\n{line}\nWarning: {e}"
                            )
                            new_line = line
                        output = [*output, new_line]
                    stmt = "\n".join(output)
                    model_file.write(stmt)

    def run(self, select: str = "", exclude: str = "", refresh: bool = False):
        """Compile the sql files then runs DBT

        Args:
            select (:obj:`str`, optional): select specifying comma-separated model names
            exclude (:obj:`str`, optional): exclude specifying comma-separated model names
            refresh (:obj:`bool`, optional): when True, refresh your incremental models
        """
        self.transpile()
        cmd = ["dbt", "run", "--fail-fast"]
        if refresh:
            cmd.append("--full-refresh")
        if select:
            cmd.append("--select")
            cmd.append(select)
        if exclude:
            cmd.append("--exclude")
            cmd.append(exclude)
        subprocess.run(cmd, env=os.environ.copy())  # nosec

    @staticmethod
    def add_omop_indexes():
        """Add OMOP indexes."""

        cmd = ["dbt", "run-operation", "add_omop_indexes"]
        subprocess.run(cmd, env=os.environ.copy())  # nosec

    def add_omop_primary_keys(self, indexes=False):
        """Add primary keys and foreign keys to the OMOP schema. Optionally add indexes.

        Args:
            indexes (:obj:`bool`, optional): specify that the operation apply indexes.
        """
        cmd = ["dbt", "run-operation", "add_omop_primary_keys"]
        subprocess.run(cmd, env=os.environ.copy())  # nosec
        if indexes:
            self.add_omop_indexes()
